# Doomsday IRC 2007
# Core IRC functions Erik Fears 2003
# Game server connection Phil "CrazySpence" Spencer 2007

 
use strict;
use Event;
use Socket;
use IO::Select;
use IO::Socket::INET;
use DBD::mysql;

require "log.pl";
require "options.pl";

#Options

my %OPTIONS = return_options();

my %IRC_FUNCTIONS = (
                     '001'     => \&m_perform,
                     'NICK'    => \&m_nick,
                     'KICK'    => \&m_kick,
                     'PART'    => \&m_part,
                     'PING'    => \&m_ping,
                     'PRIVMSG' => \&m_privmsg,
                     'QUIT'    => \&m_part,
                    );


#Global Variables
my $IRC_MYIP;        #My ip
my $IRC_SOCKET;      #IRC Connection
my $IRC_DATA;        #Data read from IRC
my $SELECT;          #IO::Select
my @DCC;             #List of DCC connections (source hashes)

my @IRC_QUEUE;       #irc message queue

### IRC FUNCTIONS
#
# Initialize IRC socket
#
sub irc_init #()
{
   if(!socket($IRC_SOCKET, PF_INET, SOCK_STREAM, getprotobyname('tcp')))
   {
      do_log(sprintf('IRC -> Error initializing IRC socket: %s', $!));
      die;
   }

   $SELECT = new IO::Select;
   $SELECT->add($$IRC_SOCKET);

   Event->timer(interval=>1, cb=>\&irc_dequeue);
}


# irc_cycle
#
# Run select() on the IRC client and dcc connections to
# check for new data. Reconnect if needed.

sub irc_cycle #()
{
   my $handle;
   my $newhandle;
   my $dcc;
   my @ready;
   my @errored;
   my $query;
  
   #do error events
   @errored = $SELECT->has_error(0);
   
   foreach $handle (@errored)
   {
      if($handle == $$IRC_SOCKET)
      {
         do_log('IRC -> IRC socket has_error');
         irc_reconnect();
         next;
      }
   }

   #do read events
   @ready = $SELECT->can_read(.1);
  
   foreach $handle (@ready)
   {
      #Data from IRC server
      if($handle == $$IRC_SOCKET)
      {
         irc_read();
         next;
      }
   }
}

# irc_connect
#
# Connect to IRC and send registration data
#

sub irc_connect #()
{
   if(!connect($$IRC_SOCKET, sockaddr_in($OPTIONS{IRC_PORT}, inet_aton($OPTIONS{IRC_HOST}))))
   {
      do_log(sprintf('IRC -> Error connecting to IRC host: %s', $!));
      irc_reconnect();
      return;   
   }

   irc_send(sprintf("NICK %s", $OPTIONS{IRC_NICK} ));
   irc_send(sprintf("USER %s %s %s :%s", $OPTIONS{IRC_NICK}, $OPTIONS{IRC_NICK},$OPTIONS{IRC_NICK},
                                         $OPTIONS{IRC_REALNAME} ));
   return;
}



# irc_reconnect
#
# Reconnct to IRC server
#

sub irc_reconnect #()
{

   do_log('IRC -> Reconnecting to server');

   close($$IRC_SOCKET);
   $SELECT->remove($$IRC_SOCKET);

   if(!socket($IRC_SOCKET, PF_INET, SOCK_STREAM, getprotobyname('tcp')))
   {
      do_log(sprintf('IRC -> Error initializing IRC socket: %s', $!));
      die;
   }

   sleep($OPTIONS{RECONNECT});
   irc_connect();
   return;
}



# irc_send
#
# Send data to IRC server
#
# $_[0] IRC Data to send

sub irc_send #($data)
{
   my $data = $_[0];

   $data .= "\n\r";

   push @IRC_QUEUE, $data;
}

sub irc_dequeue
{
   
   return if(!@IRC_QUEUE);

   if(!send($$IRC_SOCKET, shift @IRC_QUEUE, 0))
   {
      do_log(sprintf('IRC -> send() error: %s', $!));
      irc_reconnect();
   }
}


# irc_msg
#
# Send data to a user (given their source)
#

sub irc_msg # \%source, $msg
{
   my $source = $_[0];
   my $msg    = $_[1];

   my @lines;
   my $line;

   @lines = split(/[\n\r]+/, $msg);
 
   return if (!$source);

   irc_send(sprintf('NOTICE %s :%s', $$source{alias}, $msg));
   
}


# irc_channel
#
# Send data to the channel
#

sub irc_channel #\$msg
{
   my $msg = $_[0];

  irc_send( sprintf("PRIVMSG %s :%s\r\n", $OPTIONS{IRC_CHANNEL}, $msg));

}

sub global_channel #\$msg
{
  my $msg = $_[0];
  
  if($OPTIONS{GLOBALRELAY} == 0) { return; }
  irc_send( sprintf("PRIVMSG %s :%s\r\n", $OPTIONS{GLOBALCHAN}, $msg));

}   

sub guild_channel #\$msg
{
  my $msg = $_[0];
  
  if($OPTIONS{GUILDRELAY} == 0) { return; }
  irc_send( sprintf("PRIVMSG %s :%s\r\n", $OPTIONS{GUILDCHAN}, $msg));

}

# irc_read
#
# Read data from IRC server
#

sub irc_read #()
{
   my $data;
   my $pos;
   my $line;

   if(sysread($$IRC_SOCKET, $data, 512) == 0)
   {
      do_log('IRC -> Read error from server');
      irc_reconnect();
      return;
   }

   $data = $IRC_DATA . $data;

   while(($pos = index($data, "\n")) != -1)
   {
      $line = substr($data, 0, $pos + 1, "");
      chomp $line;
      irc_parse($line);
   }
   $IRC_DATA = $data;
}

# irc_parse
#
# Parse a line of data from the irc server
#

sub irc_parse #($line)
{
   my $line = $_[0];
  
   my @parv;
   my $command;
   my $message;
   my %source;
   chomp $line;

   @parv = split(/\s+/, substr($line, 0, index($line, ':', 1)));
   $message = substr($line, index($line, ':', 1) + 1, length($line)); 

   push @parv, $message;

   if($parv[0] =~ /:/)
   {
      $parv[0] = substr($parv[0], 1, length($parv[0]));
   }
   else
   {
      unshift @parv, $OPTIONS{IRC_HOST};
   }

   #parse the nick!user@host if it exists
   if($parv[0] =~ /([^!]+)!([^@]+)@(.*)/)
   {
      $source{nickname} = $1;
      $source{alias} = $1;
      $source{username} = $2;
      $source{hostname} = $3;
      $source{is_user}  = 1;
   }
   else { $source{is_user}   = 0; }
  
   
   if(exists($IRC_FUNCTIONS{$parv[1]}))
   {
      $IRC_FUNCTIONS{$parv[1]}(\@parv, \%source);
   }
}


# m_ping
#
# PING from server. 
#
# parv[0] = SOURCE
# parv[1] = PING
# parv[2] = PACKAGE
#

sub m_ping # \@parv, \%source
{
   my $parv = $_[0];
   irc_send(sprintf('PONG :%s', $$parv[2]));  
} 



# m_perform
#
# Successfull connection (perform)
#

sub m_perform # \@parv, \%source
{
   my $parv = $_[0];

   irc_send(sprintf('JOIN %s', $OPTIONS{IRC_CHANNEL}));

   #our IP should be in the 001
   if($$parv[3] =~ /@([^\s]+)/)
   {
    $IRC_MYIP = unpack("N",inet_aton($1));
   }
   
   else { $IRC_MYIP = unpack("N",inet_aton($OPTIONS{LOCALADDRESS}));
  #die "Did not get IP from IRC server (001)!"; 
  }
}


# m_privmsg
#
# privmsg to channel OR user
#
# parv[0] source
# parv[1] PRIVMSG
# parv[2] target
# parv[3] message

sub m_privmsg #\@parv, \%source
{
   my $parv = $_[0];
   my $source = $_[1];

   if($$parv[2] =~ /#/) { 
      #Said in channel
      if(($$parv[2] eq $OPTIONS{GUILDCHAN}) && ($OPTIONS{GUILDRELAYOUT} == 1)) {
         if($$parv[3] =~ /^!/) {
            $$parv[3] = substr($$parv[3], 1, length($$parv[3]));
            outbound_relay($$source{nickname},$$parv[3],"GUILD");   
         }
      } elsif(($$parv[2] eq $OPTIONS{GLOBALCHAN}) && ($OPTIONS{GLOBALRELAYOUT} == 1)) {
         if($$parv[3] =~ /^!/) {
            $$parv[3] = substr($$parv[3], 1, length($$parv[3]));
            outbound_relay($$source{nickname},$$parv[3],"RELAY");   
         }   
      }
   } else {
      #Said in private
      if($OPTIONS{GLOBALRELAYOUT} == 1) {
         outbound_relay($$source{nickname},$$parv[3],"RELAY");   
      } else {
         irc_msg($source,"Private messages unavailable for relay please use ! in the channel");
      }
   }
}

# m_nick
#
# User changed nick
#
# parv[0] source
# parv[1] NICK
# parv[2] target nick 

sub m_nick #\@parv, \%source
{
   #No longer used
}


# m_part
#
# user parted or quit
#
# parv[0] source
# parv[1] QUIT/PART
# parv[2] message

sub m_part
{
   #No Longer Used
}

# m_kick
#
# user got kicked
#
# parv[0] source
# parv[1] KICK
# parv[2] channel
# parv[3] user kicked
# parv[4] kick message

sub m_kick
{
   #No Longer Used
}

return 1;
