#!/usr/bin/perl
#VO to IRC relay server

use strict;
use Event;
use Socket;
use IO::Select;
use IO::Socket::INET;
use DBD::mysql;

require "log.pl";
require "options.pl";
require "irccls.pl";

my %OPTIONS = return_options();

my $PASSWORD = $OPTIONS{PASSWORD};
my $RELAY    = 0; #Have a relayer variable;
my $SOCKET; #Main Socket
my $SELECT; 
my $MINUTE; #minute timer
my @CPOOL;  #Connection pool hashes
my $SQL;    #SQL Connection;
main();

sub main()
{
    do_log(sprintf('MAIN -> Relay Build %s', $OPTIONS{DD_BUILD}));
    if (!$OPTIONS{DEBUG})
    {
        fork and exit;
    }
    if ($OPTIONS{DEBUG}) { do_log("MAIN -> Initializing TCP Sockets") }
    server_init(); #Start the Relay server
    if ($OPTIONS{DEBUG}) { do_log("MAIN -> Starting Timers ") }
    relay_init(); #start timers
    if ($OPTIONS{DEBUG}) { do_log("MAIN -> Connecting to IRC ") }
    irc_init(); #Start IRC connection
    irc_connect();

    while (1) 
    {
        server_cycle();
        irc_cycle();
        Event::sweep();
    }
}

sub relay_init {
    my $time;
    my $data_source;
	
    $time = time + (60 - (time % 60));
    $MINUTE = Event->timer(at=>$time, interval=>60,hard=>1,cb=>\&relay_minute);

    if($OPTIONS{DB_LOG} ==1) {
       #Connect to database
       if ($OPTIONS{DEBUG}) { do_log("RELAY INIT -> Connecting to MySQL Database ") }
       $data_source = sprintf('DBI:mysql:database=%s;host=%s;port=%d', $OPTIONS{DB_DB}, $OPTIONS{DB_HOST}, $OPTIONS{DB_PORT});
  
       $SQL = DBI->connect($data_source, $OPTIONS{DB_USER}, $OPTIONS{DB_PASS});
       die if(!$SQL);
    }
}

sub relay_minute {
    #Ping all connected clients
    my $pool;
  
    foreach $pool (@CPOOL) {
       if ( (time - $$pool{ping}) > 180 ) {
           #no response in 3 minutes
	    server_cleanup($$pool{handle});
       } else {
	    player_msg($pool,"PING");
           if($RELAY == 0) {
               $$pool{relay} = 1;
               player_msg($pool,"ACTIVATED");
               irc_channel(sprintf("%s Registered as relay",$$pool{nickname}));
               $RELAY = $$pool{handle}; 
           }
       }	
    }
}

sub inbound_relay {
    my $source  = $_[0];
    my $message = $_[1];
    my $type    = $_[2];
	
    if ($OPTIONS{DEBUG}) { do_log(sprintf("RELAY -> %s",$message)) }
    if ($type eq "RELAY")  {
        global_channel($message);
        if($OPTIONS{DB_LOG} == 1) { sql_log($message,"RELAY") }
    } elsif ($type eq "GUILD") {
        guild_channel($message);
        if($OPTIONS{DB_LOG} == 1) { sql_log($message,"GUILD") }
    }
}

sub outbound_relay {
    my $handle  = $RELAY;
    my $message = $_[1];
    my $outbound;
    my $nickname = $_[0];
    my $type = $_[2];

    if ($RELAY == 0) { return; } #No relayer available;
    
    if ($type eq "RELAY" ) {
       if ( $message =~ /^\/me / ) {
          $outbound = sprintf("[irc] * %s %s",$nickname,substr($message, 4, length($message)));
          if (!send($handle, "RELAY " . $outbound . "\n", 0))
          {
             if ($OPTIONS{DEBUG}) { do_log("ERROR -> SEND failed on relay socket closing connection\n") }
             server_cleanup($RELAY);
          } else {
              if($OPTIONS{DB_LOG} == 1) { sql_log($outbound,"RELAY") }          
          }       
       } else {
          $outbound = sprintf("[irc] <%s> %s",$nickname,$message);
          if (!send($handle, "RELAY " . $outbound . "\n", 0))
          {
             if ($OPTIONS{DEBUG}) { do_log("ERROR -> SEND failed on relay socket closing connection\n") }
             server_cleanup($RELAY);
          } else {
             if($OPTIONS{DB_LOG} == 1) { sql_log($outbound,"RELAY") }          
          }
       }
    }
    if ($type eq "GUILD" ) {
      if ( $message =~  /^online/ ) {
          if (!send($handle, "ONLINE\n", 0))
          {
             if ($OPTIONS{DEBUG}) { do_log("ERROR -> SEND failed on relay socket closing connection\n") }
             server_cleanup($RELAY);
          }        
      } elsif ( $message =~ /^motd/ ) {
          if (!send($handle, "MOTD\n", 0))
          {
             if ($OPTIONS{DEBUG}) { do_log("ERROR -> SEND failed on relay socket closing connection\n") }
             server_cleanup($RELAY);
          }         
      } elsif ( $message =~ /^bank/ ) {
          if (!send($handle, "BANK\n", 0))
          {
             if ($OPTIONS{DEBUG}) { do_log("ERROR -> SEND failed on relay socket closing connection\n") }
             server_cleanup($RELAY);
          }         
      } elsif ( $message =~ /^score/ ) {
          if (!send($handle, "SCORE\n", 0))
          {
             if ($OPTIONS{DEBUG}) { do_log("ERROR -> SEND failed on relay socket closing connection\n") }
             server_cleanup($RELAY);
          }         
      } elsif ( $message =~ /^\/me / ) {
          $outbound = sprintf("[irc] * %s %s",$nickname,substr($message, 4, length($message)));      
          if (!send($handle, "GUILD " . $outbound . "\n", 0))
          {
             if ($OPTIONS{DEBUG}) { do_log("ERROR -> SEND failed on relay socket closing connection\n") }
             server_cleanup($RELAY);
          } else {
             if($OPTIONS{DB_LOG} == 1) { sql_log($outbound,"GUILD") }          
          }      
       } else {
          $outbound = sprintf("[irc] <%s> %s",$nickname,$message);
          if (!send($handle, "GUILD " . $outbound . "\n", 0))
          {
             if ($OPTIONS{DEBUG}) { do_log("ERROR -> SEND failed on relay socket closing connection\n") }
             server_cleanup($RELAY);
          } else {
             if($OPTIONS{DB_LOG} == 1) { sql_log($outbound,"GUILD") }          
          }
       }
   }
}


sub server_init() 
{
    $SELECT = new IO::Select;
    $SOCKET = new IO::Socket::INET( Proto     => "tcp",
                                  Listen    => 100,
                                  LocalPort => $OPTIONS{RELAYPORT},
                                  Reuse     => "1"
                                );
    die "Could not create socket: $!\n" unless $SOCKET;
    $SELECT->add($SOCKET);
}

sub server_cycle()
{
    my $NEW_CONNECTION;
    my @ready;
    my $handle;
    my $data;
  
    @ready = $SELECT->can_read(.1);
  
    foreach $handle (@ready)
    {
        if ($handle == $SOCKET) #new connection time
        {
            $NEW_CONNECTION = $SOCKET->accept();
            $SELECT->add($NEW_CONNECTION);
        } else
        {
            if(sysread($handle, $data, 512) == 0) #read it or close it
            {
                #report error and cleanup any players associated with that connection 
                if ($OPTIONS{DEBUG}) { do_log("ERROR -> READ failed on socket closing connection\n") }
                server_cleanup($handle);
            } else {
                server_recieve($handle,$data);
            }
        }
    }
}

sub server_recieve #\$handle,$data
{
    my $handle = $_[0];
    my $data = $_[1];
    my @message;
    my %source;
    my $pool;
   
    @message = split(/\s+/,$data);
    if ($message[0] eq "REGISTER") {
        @message = split(/["^"]/,$data); #Account for vo names with spaces
        if ($message[1] && !is_registered($message[1]) && (ltrim($message[2]) eq $PASSWORD)) {
            $source{handle} = $handle;
            $source{nickname} = $message[1];
            if($RELAY == 0) {
                $source{relay} = 1;
                player_msg(\%source,"ACTIVATED");
                irc_channel(sprintf("%s Registered as relay",$message[1]));
                $RELAY = $source{handle};
            }
            register_player(\%source);
        } else {
	     send($handle,"ERROR: Registration failed due to nickname error\n",0);
       }
          	
   }
   if ($message[0] eq "RELAY") { #Message from general chat recieved
       if (getsource($handle)) {
           inbound_relay(\%source,substr($data,(index($data, "",length("RELAY ")))),"RELAY");  	
       } else {
	    send($handle,"ERROR: You are not Registered\n",0);
	    if ($OPTIONS{DEBUG}) { do_log("ERROR -> Unregistered player sending RELAY, disconnecting") }
	    server_cleanup($handle);   
       }
   }
   if ($message[0] eq "GUILD") { #Message from general chat recieved
       if (getsource($handle)) {
          inbound_relay(\%source,substr($data,(index($data, "",length("GUILD ")))),"GUILD");  	
       } else {
           send($handle,"ERROR: You are not Registered\n",0);
	    if ($OPTIONS{DEBUG}) { do_log("ERROR -> Unregistered player sending RELAY, disconnecting") }
	    server_cleanup($handle);   
       }
   }

   if ($message[0] eq "LOGOUT") {
       server_cleanup($handle);   	
   }
   if ($message[0] eq "PONG") {
       server_pong(getsource($handle));	
   }
}

sub server_pong {
    my $source = $_[0];

    $$source{ping} = time;
}

sub server_cleanup #\$handle
{
    #if a client disconnects compare its handle against players and remove any
    #matching players from the CPOOL
    my $handle = $_[0];
    my @DISCARD;
    my $discard;
    my $pool;

    foreach $pool (@CPOOL)
    {
        if($handle == $$pool{handle}) {
            #Originally I had unregister_player here but that changes the length of
            #@CPOOL and then players are missed so I created a DISCARD array for everyone
            #I need to get rid of.
            push @DISCARD, $pool;
        }
    }
    foreach $discard (@DISCARD) {
        unregister_player($discard); #Good bye!
    }
    $SELECT->remove($handle);
    $handle->close;
}

sub player_msg # \%source, $msg
{
    my $source = $_[0];
    my $msg    = $_[1];
    my $nickname;
    my @lines;
    my $line;
    my $handle;
   
    @lines = split(/[\n\r]+/, $msg);
 
    return if (!$source);

    if($$source{handle})
    {
        foreach $line (@lines)
        {
            if ($OPTIONS{DEBUG}) { do_log(sprintf("SEND %s -> %s", $$source{nickname}, $line)) }
            if (!send($$source{handle}, $line . "\n", 0))
            {
                if ($OPTIONS{DEBUG}) { do_log("ERROR -> SEND failed on socket closing connection\n") }
                server_cleanup($$source{handle});
            }
      }
   }
}   

sub global_msg #\$msg
{
    my $msg    = $_[0];
    my $handle;
    my @handles;
   
    @handles = $SELECT->handles;
    foreach $handle (@handles) {
        send($handle, "*** " . $msg . "\n", 0);
    }
} 

sub register_player #\%source
{
   #old DD had irc_dcc which set up the connection and added them to a searchable hash
   #but this incarnation of dd the connection is already made so We just need to add $source
   #to @cpool so when needed we can find this player later
   my $source = $_[0];
    
   $$source{ping} = time;
   push @CPOOL, $source;
   player_msg($source,"Logged In.");
   #show_cpool();
}  

sub unregister_player #\%source
{
    #take leaving players out of the hash pool
    #This does not disconnect the player as the player may be connected to a hub
    #client that accepts multiple connections

    my $source = $_[0];
    my $i;

    return if (!$source);

    for(my $i = 0; $i < scalar @CPOOL; $i++)
    {
        if($CPOOL[$i] == $source)
        {
            if ($OPTIONS{DEBUG}) { do_log(sprintf("MAIN -> %s logging off",$$source{nickname})) }
            if($$source{relay} == 1) {
                $RELAY = 0;
                irc_channel(sprintf("Relayer %s disconnected",$$source{nickname}));
            }
            splice(@CPOOL, $i, 1);
         
       }
    }
    #show_cpool();
    return;
}

sub show_cpool
{
    #displays nicks in cpool so I can tell if it is deleting items properly
    my $pool;
    foreach $pool (@CPOOL)
    {
        printf("connection: %s\n",$$pool{nickname});
    }
    return;
}

sub getsource #\$id
{
    my $handle = $_[0];
    my $pool;

    foreach $pool (@CPOOL)
    {
        if($$pool{handle} == $handle)
        {
            return $pool;
        }
    }
    return 0;
}

sub is_registered #\$nickname
{
    my $nickname = $_[0];
    my $pool;

    foreach $pool (@CPOOL)
    {
        if(lc($$pool{nickname}) eq lc($nickname))
        {
            return 1;
        }
    }
    return 0;
}

sub ltrim($)
{
    my $string = shift;
    $string =~ s/^\s+//;
    return $string;
}

sub sql_log
{
   #Add text to SQL log
   my $data = $_[0];
   my $type = $_[1];
   my $query;

   $query = $SQL->prepare("INSERT INTO log SET data=?,type=?");
   $query->execute($data,$type);
    
}